import os
from typing import Optional

from gql import gql, Client
from graphql import DocumentNode
from gql.transport.requests import RequestsHTTPTransport

from logging import getLogger

log = getLogger(__name__)


class Spacelift:
    def __init__(
        self,
        base_url: Optional[str] = None,
        key_id: Optional[str] = None,
        key_secret: Optional[str] = None,
    ):
        params = self._validate_params(base_url, key_id, key_secret)
        self.base_url = params["base_url"]
        self.key_id = params["key_id"]
        self.key_secret = params["key_secret"]
        self._jwt = None
        self.headers = {}
        self.transport = RequestsHTTPTransport(url=self.base_url, headers=self.headers)
        self.client = Client(transport=self.transport)

    @staticmethod
    def _validate_params(
        base_url: Optional[str] = None,
        key_id: Optional[str] = None,
        key_secret: Optional[str] = None,
    ) -> dict:
        """
        Ensures each parameter is provided by either the constructor or environment variables.
        TODO: add support for spacelift's CLI config file, etc.
        """
        base_url = base_url or os.environ.get("SPACELIFT_BASE_URL")
        key_id = key_id or os.environ.get("SPACELIFT_KEY_ID")
        key_secret = key_secret or os.environ.get("SPACELIFT_KEY_SECRET")
        result = {
            "base_url": base_url,
            "key_id": key_id,
            "key_secret": key_secret,
        }
        null_params = [k for k, v in result.items() if not v]
        if null_params:
            raise ValueError(f"Missing spacelift params: {null_params}")
        return result

    def _execute(
        self, query: DocumentNode, variable_values: Optional[dict] = None
    ) -> dict:
        self.transport.headers["Authorization"] = f"Bearer {self.jwt}"
        return self.client.execute(query, variable_values=variable_values)

    def _get_jwt(self):
        query_text = f"""
        mutation GetSpaceliftToken($keyId: ID!, $keySecret: String!) {{
            apiKeyUser(id: $keyId, secret: $keySecret) {{
                id
                jwt
            }}
        }}
        """
        variable_values = {
            "keyId": self.key_id,
            "keySecret": self.key_secret,
        }
        query = gql(query_text)
        result = self.client.execute(
            query, variable_values=variable_values
        )  # cannot use self._execute() here because it would cause infinite recursion
        log.debug(f"get_jwt result: {result}")
        if result.get("apiKeyUser") is None:
            raise ValueError(f"Invalid spacelift keyId: {self.key_id}")
        return result["apiKeyUser"]["jwt"]

    @property
    def jwt(self):
        if not self._jwt:
            self._jwt = self._get_jwt()
        return self._jwt

    def get_stacks(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "space"]
        query_text = f'query Stacks {{ stacks {{ {" ".join(query_fields)}  }} }}'
        query = gql(query_text)
        return self._execute(query)["stacks"]

    def get_stack_by_id(
        self, stack_id: str, query_fields: Optional[list[str]] = None
    ) -> dict:
        if query_fields is None:
            query_fields = ["id", "space"]
        query_text = f'query Stack($id: ID!) {{ stack(id: $id) {{ {" ".join(query_fields)}  }} }}'
        variable_values = {
            "id": stack_id,
        }
        query = gql(query_text)
        return self._execute(query, variable_values=variable_values)["stack"]

    def get_spaces(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "name"]
        query_text = f'query Spaces {{ spaces {{ {" ".join(query_fields)}  }} }}'
        query = gql(query_text)
        return self._execute(query)["spaces"]

    def get_space_by_id(
        self, space_id: str, query_fields: Optional[list[str]] = None
    ) -> dict:
        if query_fields is None:
            query_fields = ["id", "name"]
        query_text = f'query Space($id: ID!) {{ space(id: $id) {{ {" ".join(query_fields)}  }} }}'
        variable_values = {
            "id": space_id,
        }
        query = gql(query_text)
        return self._execute(query, variable_values=variable_values)["space"]

    def get_contexts(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "name"]
        query_text = f'query Contexts {{ contexts {{ {" ".join(query_fields)}  }} }}'
        query = gql(query_text)
        return self._execute(query)["contexts"]

    def get_context_by_id(
        self, context_id: str, query_fields: Optional[list[str]] = None
    ) -> dict:
        if query_fields is None:
            query_fields = [
                "id",
                "name",
            ]
        query_text = f'query Context($id: ID!) {{ context(id: $id) {{ {" ".join(query_fields)}  }} }}'
        variable_values = {
            "id": context_id,
        }
        query = gql(query_text)
        return self._execute(query, variable_values=variable_values)["context"]

    def create_context(
        self,
        context_name: str,
        space_id: str,
        description: str = "",
        labels: list[str] = None,
        envvars: list[dict] = None,
    ):
        if labels is None:
            labels = []
        if envvars is None:
            envvars = []

        query_text = f"""
        mutation ContextCreateV2($name: String!, $envvars: [ConfigInput!], $space: ID, $description: String, $labels: [String!]) {{
            contextCreateV2(input: {{name: $name, configAttachments: $envvars, space: $space, description: $description, labels: $labels }}) {{
                id
                name
                config {{
                    id
                    value
                }}
            }}
        }}
        """

        variable_values = {
            "name": context_name,
            "description": description,
            "labels": labels,
            "space": space_id,
            "envvars": envvars,
        }

        query = gql(query_text)
        result = self._execute(query, variable_values=variable_values)
        log.debug(f"create_context result: {result}")
        return result

    def delete_context(self, context_id: str):
        query_text = f"""
        mutation ContextDelete($id: ID!) {{
            contextDelete(id: $id) {{
                id
            }}
        }}
        """
        variable_values = {
            "id": context_id,
        }
        query = gql(query_text)
        result = self._execute(query, variable_values=variable_values)
        log.debug(f"delete_context result: {result}")
        return result

    def create_space(
        self,
        space_name: str,
        parent_space_id: str,
        description: str,
        labels: list[str] = None,
        inherit_entities: bool = True,
    ):
        if labels is None:
            labels = []

        query_text = f"""
        mutation SpaceCreate($name: String!, $parent: ID!, $description: String!, $labels: [String!], $inheritEntities: Boolean!) {{
            spaceCreate(input: {{name: $name, parentSpace: $parent, description: $description, labels: $labels, inheritEntities: $inheritEntities}}) {{
                id
                name
                description
                labels
            }}
        }}
        """
        variable_values = {
            "name": space_name,
            "parent": parent_space_id,
            "description": description,
            "labels": labels,
            "inheritEntities": inherit_entities,
        }
        query = gql(query_text)
        result = self._execute(query, variable_values=variable_values)
        log.debug(f"create_space result: {result}")
        return result

    def delete_space(self, space_id: str):
        query_text = f"""
        mutation SpaceDelete($id: ID!) {{
            spaceDelete(space: $id) {{
                id
            }}
        }}
        """
        variable_values = {
            "id": space_id,
        }
        query = gql(query_text)
        result = self._execute(query, variable_values=variable_values)
        log.debug(f"delete_space result: {result}")
        return result

    def trigger_run(self, stack_id: str, query_fields: Optional[list[str]] = None):
        if query_fields is None:
            query_fields = ["id", "branch"]

        query_text = f"""
        mutation RunTrigger($stack: ID!) {{
            runTrigger(stack: $stack, runType: PROPOSED) {{
                {" ".join(query_fields)}
            }}
        }}
        """
        variable_values = {
            "stack": stack_id,
        }
        query = gql(query_text)
        result = self._execute(query, variable_values=variable_values)["runTrigger"]
        log.debug(f"trigger_run result: {result}")
        return result

    def get_blueprints(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "name"]
        newline = "\n"
        query_text = f"""
        query SearchBlueprints($input: SearchInput!) {{
         searchBlueprints(input: $input) {{ 
          edges {{
           node {{ 
            {newline.join(query_fields)}
           }}
          }} 
         }} 
        }}
        """
        variable_values = {
            "input": {
                "fullTextSearch": "",
                "predicates": [],
            }
        }
        query = gql(query_text)
        return self._execute(query, variable_values=variable_values)[
            "searchBlueprints"
        ]["edges"]

    def get_blueprint_by_id(
        self, blueprint_id: str, query_fields: Optional[list[str]] = None
    ) -> dict:
        if query_fields is None:
            query_fields = ["id", "name"]
        newline = "\n"
        query_text = f"""
        query blueprint($id: ID!) {{
         blueprint(id: $id) {{ 
          {newline.join(query_fields)}
         }} 
        }}
        """
        variable_values = {
            "id": blueprint_id,
        }
        query = gql(query_text)
        return self._execute(query, variable_values=variable_values)["blueprint"]

    def create_stack_from_blueprint(
        self, blueprint_id: str, inputs: list[dict]
    ) -> dict:
        query_text = f"""
        mutation BlueprintCreateStack($id: ID!, $input: BlueprintStackCreateInput!) {{
            blueprintCreateStack(id: $id, input: $input) {{
                stackID
                runID
            }}
        }}
        """
        variable_values = {
            "id": blueprint_id,
            "input": {"templateInputs": inputs},
        }
        query = gql(query_text)
        raw_result = self._execute(query, variable_values=variable_values)
        result = raw_result["blueprintCreateStack"]
        log.debug(f"create_stack_from_blueprint result: {result}")
        return result


def space_context_test():
    sl = Spacelift(os.environ.get("SPACELIFT_BASE_URL"))
    # result = sl.get_stacks()
    # print(result)
    # result = sl.get_stack_by_id(stack_id="mcr-tf", query_fields=["id", "name"])
    # print(result)
    #
    # result = sl.get_stacks(["id", "name", "branch", "namespace", "repository", "state"])
    # print(result)

    # result = sl.get_context_by_id(context_id="customer-a", query_fields=["id", "name", "config { id value writeOnly }"])
    LIBRARY_TEST_NAME = "library-test-customer-a"

    result = sl.get_contexts(
        query_fields=["id", "name", "config { id value writeOnly }"]
    )
    print(result[0])

    try:
        print('deleting context_id="abc"')
        result = sl.delete_context(context_id="abc")
        print(f"delete_context(context_id='abc') result: {result}")
    except Exception as e:
        print(f"exception: {type(e)}:{e}")

    try:
        result = sl.create_context(
            context_name=LIBRARY_TEST_NAME,
            space_id="root",
            envvars=[
                {
                    "id": "FOO",
                    "value": "bar",
                    "type": "ENVIRONMENT_VARIABLE",
                    "writeOnly": False,
                },
                {
                    "id": "BAZ",
                    "value": "qux",
                    "type": "ENVIRONMENT_VARIABLE",
                    "writeOnly": True,
                },
            ],
        )
        print(f"create_context result: {result}")

    except Exception as e:
        print(f"create_context exception: {type(e)}:{e}")
        result = sl.delete_context(context_id=LIBRARY_TEST_NAME)
        print(f"delete_context result: {result}")

    spaces = sl.get_spaces(query_fields=["id", "name", "description", "parentSpace"])
    print(spaces)

    delete_spaces = [space for space in spaces if space["name"] == LIBRARY_TEST_NAME]
    print(delete_spaces)

    for space in delete_spaces:
        print(space["id"])
        result = sl.delete_space(space_id=space["id"])
        print(f"delete_space result: {result}")

    if not delete_spaces:
        result = sl.create_space(
            space_name=LIBRARY_TEST_NAME,
            parent_space_id="root",
            description=f"{LIBRARY_TEST_NAME} space",
            labels=["customer", "a"],
        )

        print(f"create_space result: {result}")


def blueprint_test():
    sl = Spacelift(os.environ.get("SPACELIFT_BASE_URL"))
    # result = sl.get_blueprints(query_fields=["id", "name"])
    bp_id = "fcr-aw4-1-az2-1"
    result = sl.get_blueprint_by_id(
        blueprint_id=bp_id, query_fields=["id", "name", "inputs { id name type }"]
    )
    print(result)

    inputs = [
        {
            "id": "environment",
            "value": "dev",
        },
        {
            "id": "customer_name",
            "value": "abc",
        },
        {
            "id": "description",
            "value": "abc dev environment",
        },
        {"id": "context_id", "value": "svc_abcd-context"},
        {"id": "space_id", "value": "svc_abcd-space-01HJP3651BQT2GY2EQSG6KVWDB"},
    ]
    result = sl.create_stack_from_blueprint(blueprint_id=bp_id, inputs=inputs)
    print(result)


def main():
    # space_context_test()
    blueprint_test()


if __name__ == "__main__":
    main()



import random
import string
from typing import Optional

from spacelift import Spacelift


def field_filter(item: dict, query_fields: list[str]):
    return {field: item[field] for field in item if field in query_fields}


class MockSpacelift:
    def __init__(self):
        self.base_url = "http://test_spacelift"
        self.key_id = "abc"
        self.key_secret = "xyz"
        self._data = {"spaces": [], "stacks": [], "contexts": []}
        self._jwt = "mock_jwt"

    @property
    def jwt(self):
        return self._jwt

    def get_blueprints(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        raise NotImplementedError()

    def get_blueprint_by_name(
        self, blueprint_name: str, query_fields: Optional[list[str]] = None
    ):
        raise NotImplementedError()

    def create_stack_from_blueprint(self, blueprint_id: str, inputs: list[dict]):
        raise NotImplementedError()

    def get_stacks(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "space"]

        stacks = self._data["stacks"]
        stacks = [field_filter(stack, query_fields) for stack in stacks]
        return stacks

    def get_stack_by_id(self, stack_id: str, query_fields: Optional[list[str]] = None):
        if query_fields is None:
            query_fields = ["id", "space"]

        stacks = self.get_stacks(query_fields=query_fields)
        stacks = [stack for stack in stacks if stack["id"] == stack_id]
        if len(stacks) == 0:
            return None

        return stacks[0]

    def get_spaces(self, query_fields: Optional[list[str]] = None) -> list[dict]:
        if query_fields is None:
            query_fields = ["id", "name"]

        spaces = self._data["spaces"]
        spaces = [field_filter(space, query_fields) for space in spaces]
        return spaces

    def get_space_by_id(self, space_id: str, query_fields: Optional[list[str]] = None):
        if query_fields is None:
            query_fields = ["id", "name"]

        spaces = self.get_spaces(query_fields=query_fields)
        spaces = [space for space in spaces if space["id"] == space_id]
        if len(spaces) == 0:
            return None

        return spaces[0]

    def get_contexts(self, query_fields: Optional[list[str]] = None):
        if query_fields is None:
            query_fields = ["id", "name"]

        config_fields = []
        for field in query_fields:
            if field.startswith("config"):
                query_fields.remove(field)
                query_fields.append("config")
                config_fields = field.split(" ")[1:]
                config_fields = [
                    config_field
                    for config_field in config_fields
                    if config_field not in "{}"
                ]

        contexts = self._data["contexts"]
        contexts = [field_filter(context, query_fields) for context in contexts]

        if "config" in query_fields:
            for context in contexts:
                context["config"] = [
                    field_filter(envvar, config_fields) for envvar in context["config"]
                ]

        return contexts

    def get_context_by_id(
        self, context_id: str, query_fields: Optional[list[str]] = None
    ):
        if query_fields is None:
            query_fields = ["id", "name"]

        contexts = self.get_contexts(query_fields=query_fields)
        contexts = [context for context in contexts if context["id"] == context_id]
        if len(contexts) == 0:
            return None

        return contexts[0]

    def create_context(
        self,
        context_name: str,
        space_id: str,
        description: str = "",
        labels: list[str] = None,
        envvars: list[dict] = None,
    ):
        if labels is None:
            labels = []
        if envvars is None:
            envvars = []

        parent_space = self.get_space_by_id(space_id)
        if parent_space is None:
            raise Exception({"message": "unauthorized", "path": ["contextCreateV2"]})

        config = [
            {
                "id": envvar["id"],
                "value": envvar["value"],
                "writeOnly": False,
            }
            for envvar in envvars
        ]

        item = {
            "name": context_name,
            "id": context_name,
            "space": space_id,
            "description": description,
            "labels": labels,
            "config": config,
        }

        contexts = self._data["contexts"]
        for context in contexts:
            if context["name"] == context_name:
                raise Exception(
                    {
                        "message": "could not persist context: context with this slug already exists for this account",
                        "path": ["contextCreateV2"],
                    }
                )

        contexts.append(item)

        return {"contextCreateV2": field_filter(item, query_fields=["id", "name"])}

    def delete_context(self, context_id: str):
        contexts = self._data["contexts"]
        for context in contexts:
            if context["id"] == context_id:
                contexts.remove(context)
                return {"contextDelete": {"id": context_id}}

        return {"contextDelete": None}

    def create_space(
        self,
        space_name: str,
        parent_space_id: str,
        description: str = "",
        labels: list[str] = None,
        inherit_entities: bool = True,
    ):
        if labels is None:
            labels = []

        if space_name not in ("root", "legacy"):
            generated_id = "".join(
                random.choice(string.ascii_lowercase + string.digits) for _ in range(16)
            )
            id = f"{space_name}-{generated_id}"
        else:
            id = space_name

        parent_space = self.get_space_by_id(parent_space_id)
        if parent_space is None:
            if space_name == parent_space_id == "root":
                parent_space_id = None
                pass  # this is fine to create the root space
            else:
                raise Exception(
                    {"message": "parent space not found", "path": ["spaceCreate"]}
                )

        item = {
            "name": space_name,
            "id": id,
            "parent": parent_space_id,
            "description": description,
            "labels": labels,
            "inherit_entities": inherit_entities,
        }

        spaces = self._data["spaces"]
        spaces.append(item)

        return {"spaceCreate": field_filter(item, query_fields=["id", "name"])}

    def delete_space(self, space_id: str):
        spaces = self._data["spaces"]
        for space in spaces:
            if space["id"] == space_id:
                spaces.remove(space)
                return {"spaceDelete": {"id": space_id}}

        return {"spaceDelete": None}

    def trigger_run(self, stack_id: str, query_fields: Optional[list[str]] = None):
        raise NotImplementedError()
